/**
 * 
 */
package org.processmining.plugins.pnalignanalysis.export;

import java.util.Iterator;
import java.util.SortedSet;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.KeepInProMCache;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

/**
 * @author aadrians
 * Nov 15, 2012
 *
 */
@KeepInProMCache
@LogCreatorFromAlignment
public class NaiveLogCreator implements ILogCreatorAlgorithm {
	// create a log based on alignment
	private XConceptExtension conceptExt = XConceptExtension.instance();
	private XFactory factory = XFactoryRegistry.instance().currentDefault();

	public String toString(){
		return "Naively insert timestamps based on serialized movements ordering";
	}

	/**
	 * Extract log from alignment
	 */
	public XLog extractLog(PluginContext context, PNRepResult logReplayResult, XLog log, LogCreatorParam param) {
		try {
			// progress bar
			final Progress p;
			int repResSize = logReplayResult.size();
			int counter = 0;
			int traceClassNum = 0; // trace class that really inserted
			if (context != null) {
				p = context.getProgress();
				p.setIndeterminate(false);
				p.setMaximum(logReplayResult.size());
				p.setMinimum(0);
			} else {
				p = null;
			}

			// result
			XLog outputLog = factory.createLog();
			outputLog.getClassifiers().addAll(log.getClassifiers()); // add classifiers

			// create event class for each pair of movement + log 
			for (SyncReplayResult repRes : logReplayResult) {
				// progress bar update (only if GUI)
				if (context != null) {
					if (counter % 100 == 0) {
						context.log("Constructing trace class " + counter + "/" + repResSize + "...");
					}
					counter++;
					p.inc();
				}

				// filter based on fitness (if exists)
				Double fitnessVal = repRes.getInfo().get(PNRepResult.TRACEFITNESS);
				if (fitnessVal != null) {
					if (Double.compare(fitnessVal, param.getMinFitnessValue()) < 0) {
						continue;
					}
					if (Double.compare(fitnessVal, param.getMaxFitnessValue()) > 0) {
						continue;
					}
				}
				

				// store all result
				traceClassNum++;

				SortedSet<Integer> traceIndexes = repRes.getTraceIndex();
				XTrace[] newTraces = new XTrace[traceIndexes.size()]; // new constructed traces

				constructTraceWithoutDerivingTimestamp(log, traceIndexes, newTraces, repRes, param);
				for (int i = 0; i < newTraces.length; i++) {
					outputLog.add(newTraces[i]);
				}

			}

			// return constructed log
			if (traceClassNum > 0) {
				return outputLog;
			}
		} catch (Exception exc) {
			// do nothing
			exc.printStackTrace();
		}
		// if not succeed
		if (context != null) {
			context.getFutureResult(0).cancel(true);
			context.log("Log is not created. There might be no traces created or there is no net associated to the replay result found.");
		}
		return null;
	}


	/**
	 * Construct sequence of events, not extracting timestamp unless it already
	 * exists in the log
	 * 
	 * @param log
	 * @param traceIndexes
	 * @param result
	 * @param repRes
	 * @param classifier
	 */
	private void constructTraceWithoutDerivingTimestamp(XLog log, SortedSet<Integer> traceIndexes, XTrace[] result,
			SyncReplayResult repRes, LogCreatorParam param) {
		int traceCounter = 0;

		// get classifier of the original log
		XEventClassifier classifier = param.getEventClassifier();
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);
		XEventClasses evClasses = logInfo.getEventClasses();

		// dummy event class and mapping 
		TransEvClassMapping mapping = param.getMapping();

		for (int traceIndex : traceIndexes) {
			XTrace trace = log.get(traceIndex);
			Iterator<Object> nodeInstIt = repRes.getNodeInstance().iterator();

			// latest timestamp
			result[traceCounter] = factory.createTrace();
			XAttribute traceNameAttr = factory.createAttributeLiteral("concept:name", conceptExt.extractName(trace),
					conceptExt);
			Iterator<XEvent> evtIt = trace.iterator();

			result[traceCounter].getAttributes().put("concept:name", traceNameAttr);

			for (StepTypes stepType : repRes.getStepTypes()) {
				switch (stepType) {
					case L : // move on log only
						if (param.isCreateEventsForMoveLog()) {
							result[traceCounter].add(constructEventFromEvent((XEvent) evtIt.next().clone(), param,
									evClasses, "[L]"));
						}
						nodeInstIt.next();
						break;
					case LMGOOD : // synchronous move
						result[traceCounter].add(constructEventFromEvent((XEvent) evtIt.next().clone(), param,
								evClasses, "[L+M]"));
						nodeInstIt.next();
						break;
					case MREAL : // move on model
						Transition t1 = (Transition) nodeInstIt.next();
						if (param.isCreateEventsForMoveModel()) {
							// check is it mapped
							if ((mapping == null) || (mapping.get(t1) == null)
									|| (mapping.get(t1).equals(mapping.getDummyEventClass()))) {
								if (!param.isCreateEventClassForUnmappedTransitions()) {
									continue;
								}
							}
							result[traceCounter].add(constructEventFromTrans(t1, param, evClasses, "[M]"));
						}
						break;
					default : // move on model (invi)
						Transition t2 = (Transition) nodeInstIt.next();
						if (param.isCreateEventClassForUnmappedTransitions()) {
							if (param.isCreateEventsForMoveModel()) {
								result[traceCounter].add(constructEventFromTrans(t2, param, evClasses, "[M(inv)]"));
							}
						}
						break;
				}
			}
			traceCounter++;
		}
	}


	/**
	 * construct new event class that represent move log event class
	 * 
	 * @param clone
	 * @param evClass2EvClass
	 * @param param
	 * @param evClasses
	 * @param stepType
	 * @return
	 */
	private XEvent constructEventFromEvent(XEvent clone, LogCreatorParam param, XEventClasses evClasses, String stepType) {
		if (param.isCreateNewEventClassesBasedOnMovements()) {
			// create new event class with the same name, but add [move log] or [move model]
			XEventClass origEvClass = evClasses.getClassOf(clone);

			XAttribute name = factory.createAttributeLiteral("concept:name", origEvClass + stepType, this.conceptExt);
			clone.getAttributes().put("concept:name", name);
			return clone;
		} else {
			return clone;
		}
	}

	/**
	 * Construct event from transitions if there is no need to create new event
	 * class based on movement, but such transition is unmapped, a new event
	 * class is created
	 * 
	 * @param trans
	 * @param param
	 * @param stepTypeStr
	 * @return
	 */
	private synchronized XEvent constructEventFromTrans(Transition trans, LogCreatorParam param,
			XEventClasses evClasses, String stepTypeStr) {
		XEvent outputEvent = factory.createEvent();

		if (param.isCreateNewEventClassesBasedOnMovements()) {
			XAttribute name = factory.createAttributeLiteral("concept:name", trans.toString() + stepTypeStr,
					this.conceptExt);
			outputEvent.getAttributes().put("concept:name", name);
		} else {
			// see if this transition is mapped to activities
			XAttribute name = factory.createAttributeLiteral("concept:name", trans.toString() + stepTypeStr,
					this.conceptExt);
			outputEvent.getAttributes().put("concept:name", name);
		}

		XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete",
				XConceptExtension.instance());
		outputEvent.getAttributes().put("lifecycle:transition", transition);

		return outputEvent;
	}

	public boolean isSupportCreateNewEventClassBasedOnMovements() {
		return true;
	}

	public boolean isSupportCreateNewEventClassForUnmappedTransitions() {
		return true;
	}
}
